@@ -48,7 +48,6 @@ _AUTO_SIGNED_NO = 0
 _AUTO_SIGNED_YES = 1
 _AUTO_SIGNED_PASSTHROUGH = 2
 _SUBJECT_OID_DICT = {
-  # pylint: disable=bad-whitespace
   'C' : x509.oid.NameOID.COUNTRY_NAME,
   'O' : x509.oid.NameOID.ORGANIZATION_NAME,
   'OU': x509.oid.NameOID.ORGANIZATIONAL_UNIT_NAME,
@@ -57,7 +56,6 @@ _SUBJECT_OID_DICT = {
   'L' : x509.oid.NameOID.LOCALITY_NAME,
   'SN': x509.oid.NameOID.SURNAME,
   'GN': x509.oid.NameOID.GIVEN_NAME,
-  # pylint: enable=bad-whitespace
 }
 _BACKUP_MAGIC = b'caucase\0'
 _CONFIG_NAME_AUTO_SIGN_CSR_AMOUNT = 'auto_sign_csr_amount'
@@ -153,6 +151,7 @@ class CertificateAuthority(object):
       When given with a true value, auto_sign_csr_amount is stored and the
       value given on later instanciation will be ignored.
     """
+    self._current_crl_dict = {}
     self._storage = storage
     self._ca_renewal_lock = threading.Lock()
     if lock_auto_sign_csr_amount:
@@ -200,17 +199,40 @@ class CertificateAuthority(object):
     return list(self._digest_list)
 
   def _loadCAKeyPairList(self):
+    digest = self._digest_list[0]
     ca_key_pair_list = []
+    ca_certificate_chain = []
+    previous_crt_pem = None
+    previous_key = None
     for pem_key_pair in self._storage.getCAKeyPairList():
       utils.validateCertAndKey(
         pem_key_pair['crt_pem'],
         pem_key_pair['key_pem'],
       )
+      crt_pem = pem_key_pair['crt_pem']
+      crt = utils.load_ca_certificate(pem_key_pair['crt_pem'])
+      key = utils.load_privatekey(pem_key_pair['key_pem'])
       ca_key_pair_list.append({
-        'crt': utils.load_ca_certificate(pem_key_pair['crt_pem']),
-        'key': utils.load_privatekey(pem_key_pair['key_pem']),
+        'crt': crt,
+        'key': key,
+        'authority_key_identifier': utils.getAuthorityKeyIdentifier(crt),
       })
+      if previous_key is not None:
+        ca_certificate_chain.append(utils.wrap(
+          {
+            'old_pem': utils.toUnicode(previous_crt_pem),
+            'new_pem': utils.toUnicode(crt_pem),
+          },
+          previous_key,
+          digest,
+        ))
+      previous_crt_pem = crt_pem
+      previous_key = key
+    self._current_crl_dict.clear()
     self._ca_key_pairs_list = ca_key_pair_list
+    self._ca_certificate_chain = tuple(
+      ca_certificate_chain
+    )
 
   def getCertificateSigningRequest(self, csr_id):
     """
@@ -323,11 +345,9 @@ class CertificateAuthority(object):
           critical=False, # "MUST mark this extension as non-critical"
         ),
         Extension(
-          x509.AuthorityKeyIdentifier.from_issuer_subject_key_identifier(
-            ca_crt.extensions.get_extension_for_class(
-              x509.SubjectKeyIdentifier,
-            ).value,
-          ),
+          ca_crt.extensions.get_extension_for_class(
+            x509.AuthorityKeyIdentifier,
+          ).value,
           critical=False, # "MUST mark this extension as non-critical"
         ),
       ],
@@ -337,7 +357,13 @@ class CertificateAuthority(object):
         x509.CRLDistributionPoints([
           x509.DistributionPoint(
             full_name=[
-              x509.UniformResourceIdentifier(self._crl_base_url),
+              x509.UniformResourceIdentifier(
+                self._crl_base_url + (
+                  '/%i' % (
+                    utils.getAuthorityKeyIdentifier(ca_crt),
+                  )
+                ),
+              ),
             ],
             relative_name=None,
             crl_issuer=None,
@@ -361,7 +387,6 @@ class CertificateAuthority(object):
         encipher_only = decipher_only = False
       builder = builder.add_extension(
         x509.KeyUsage(
-          # pylint: disable=bad-whitespace
           digital_signature =key_usage.digital_signature,
           content_commitment=key_usage.content_commitment,
           key_encipherment  =key_usage.key_encipherment,
@@ -371,7 +396,6 @@ class CertificateAuthority(object):
           crl_sign          =False,
           encipher_only     =encipher_only,
           decipher_only     =decipher_only,
-          # pylint: enable=bad-whitespace
         ),
         # "SHOULD mark this extension critical"
         critical=key_usage_extension.critical,
@@ -509,7 +533,6 @@ class CertificateAuthority(object):
             ),
             Extension(
               x509.KeyUsage(
-                # pylint: disable=bad-whitespace
                 digital_signature =False,
                 content_commitment=False,
                 key_encipherment  =False,
@@ -519,7 +542,6 @@ class CertificateAuthority(object):
                 crl_sign          =True,
                 encipher_only     =False,
                 decipher_only     =False,
-                # pylint: enable=bad-whitespace
               ),
               critical=True, # "SHOULD mark this extension critical"
             ),
@@ -598,9 +620,15 @@ class CertificateAuthority(object):
 
   def getCACertificate(self):
     """
-    Return current CA certificate, PEM-encoded.
+    Return oldest still-valid CA certificate, PEM-encoded.
     """
-    return utils.dump_certificate(self._getCurrentCAKeypair()['crt'])
+    self._renewCAIfNeeded()
+    # XXX: _loadCAKeyPairList seems a bit too expensive for such simple operation.
+    # So do out-of-storage self._ca_key_pairs_list pruning only.
+    now = datetime.datetime.utcnow()
+    while self._ca_key_pairs_list[0]['crt'].not_valid_after <= now:
+      self._ca_key_pairs_list.pop(0)
+    return utils.dump_certificate(self._ca_key_pairs_list[0]['crt'])
 
   def getCACertificateList(self):
     """
@@ -636,24 +664,7 @@ class CertificateAuthority(object):
     purposes.
     """
     self._renewCAIfNeeded()
-    result = []
-    iter_key_pair = iter(self._ca_key_pairs_list)
-    first_key_pair = next(iter_key_pair)
-    previous_crt_pem = utils.dump_certificate(first_key_pair['crt'])
-    previous_key = first_key_pair['key']
-    for key_pair in iter_key_pair:
-      current_crt_pem = utils.dump_certificate(key_pair['crt'])
-      result.append(utils.wrap(
-        {
-          'old_pem': utils.toUnicode(previous_crt_pem),
-          'new_pem': utils.toUnicode(current_crt_pem),
-        },
-        previous_key,
-        self.digest_list[0],
-      ))
-      previous_key = key_pair['key']
-      previous_crt_pem = current_crt_pem
-    return result
+    return self._ca_certificate_chain
 
   def revoke(self, crt_pem):
     """
@@ -665,15 +676,22 @@ class CertificateAuthority(object):
     crt = utils.load_certificate(
       crt_pem,
       self.getCACertificateList(),
-      x509.load_pem_x509_crl(
-        self.getCertificateRevocationList(),
-        _cryptography_backend,
-      ),
+      crl_list=[
+        x509.load_pem_x509_crl(x, _cryptography_backend)
+        for x in self.getCertificateRevocationListDict().itervalues()
+      ],
     )
     self._storage.revoke(
       serial=crt.serial_number,
-      expiration_date=utils.datetime2timestamp(crt.not_valid_after),
+      expiration_date=utils.datetime2timestamp(
+        # https://tools.ietf.org/html/rfc5280#section-3.3
+        # An entry MUST NOT be removed
+        # from the CRL until it appears on one regularly scheduled CRL issued
+        # beyond the revoked certificate's validity period.
+        crt.not_valid_after + self._crl_life_time,
+      ),
     )
+    self._current_crl_dict.clear()
 
   def revokeSerial(self, serial):
     """
@@ -690,10 +708,13 @@ class CertificateAuthority(object):
     """
     self._storage.revoke(
       serial=serial,
-      expiration_date=utils.datetime2timestamp(max(
-        x.not_valid_after for x in self.getCACertificateList()
-      )),
+      expiration_date=utils.datetime2timestamp(
+        max(
+          x.not_valid_after for x in self.getCACertificateList()
+        ) + self._crl_life_time,
+      ),
     )
+    self._current_crl_dict.clear()
 
   def renew(self, crt_pem, csr_pem):
     """
@@ -707,10 +728,10 @@ class CertificateAuthority(object):
     crt = utils.load_certificate(
       crt_pem,
       self.getCACertificateList(),
-      x509.load_pem_x509_crl(
-        self.getCertificateRevocationList(),
-        _cryptography_backend,
-      ),
+      crl_list=[
+        x509.load_pem_x509_crl(x, _cryptography_backend)
+        for x in self.getCertificateRevocationListDict().itervalues()
+      ],
     )
     return self._createCertificate(
       csr_id=self.appendCertificateSigningRequest(
@@ -731,46 +752,82 @@ class CertificateAuthority(object):
       ),
     )
 
-  def getCertificateRevocationList(self):
+  def getCertificateRevocationListDict(self):
     """
-    Return PEM-encoded certificate revocation list.
+    Return PEM-encoded certificate revocation lists for all CAs.
     """
-    crl_pem = self._storage.getCertificateRevocationList()
-    if crl_pem is None:
-      ca_key_pair = self._getCurrentCAKeypair()
-      now = datetime.datetime.utcnow()
-      crl = x509.CertificateRevocationListBuilder(
-        issuer_name=ca_key_pair['crt'].issuer,
-        last_update=now,
-        next_update=now + self._crl_life_time,
-        extensions=[
-          Extension(
-            x509.CRLNumber(
-              self._storage.getNextCertificateRevocationListNumber(),
+    now = datetime.datetime.utcnow()
+    result = {}
+    crl_pem_dict = self._current_crl_dict
+    self._renewCAIfNeeded()
+    storage = self._storage
+    crl_number, last_update = storage.getCurrentCRLNumberAndLastUpdate()
+    has_renewed = last_update is None
+    last_update = (
+      None
+      if last_update is None else
+      utils.timestamp2datetime(last_update)
+    )
+    revoked_certificate_list = None
+    for ca_key_pair in self._ca_key_pairs_list:
+      authority_key_identifier = ca_key_pair['authority_key_identifier']
+      try:
+        crl_pem, crl_expiration_date = crl_pem_dict[authority_key_identifier]
+      except KeyError:
+        crl_pem = None
+      if crl_pem is None or crl_expiration_date < now:
+        if not has_renewed and crl_pem is not None:
+          # CRL expired, generate a new serial
+          last_update = None
+          has_renewed = True
+        if (
+          last_update is None or
+          last_update + self._crl_renew_time < now
+        ):
+          # We cannot use the existing CRL (or maybe none exist), generate a
+          # new one.
+          last_update = now
+          crl_number = storage.getNextCertificateRevocationListNumber()
+          storage.storeCRLLastUpdate(
+            last_update=utils.datetime2timestamp(last_update),
+          )
+        if revoked_certificate_list is None:
+          revoked_certificate_list = [
+            x509.RevokedCertificateBuilder(
+              serial_number=x['serial'],
+              revocation_date=utils.timestamp2datetime(x['revocation_date']),
+            ).build(_cryptography_backend)
+            for x in storage.getRevocationList()
+          ]
+        ca_crt = ca_key_pair['crt']
+        crl_pem = x509.CertificateRevocationListBuilder(
+          issuer_name=ca_crt.issuer,
+          last_update=last_update,
+          next_update=last_update + self._crl_life_time,
+          extensions=[
+            Extension(
+              x509.CRLNumber(crl_number),
+              critical=False, # "MUST mark this extension as non-critical"
             ),
-            critical=False, # "MUST mark this extension as non-critical"
-          ),
-        ],
-        revoked_certificates=[
-          x509.RevokedCertificateBuilder(
-            serial_number=x['serial'],
-            revocation_date=datetime.datetime.fromtimestamp(
-              x['revocation_date'],
+            Extension(
+              ca_crt.extensions.get_extension_for_class(
+                x509.AuthorityKeyIdentifier,
+              ).value,
+              critical=False, # No mention in RFC5280 5.2.1
             ),
-          ).build(_cryptography_backend)
-          for x in self._storage.getRevocationList()
-        ],
-      ).sign(
-        private_key=ca_key_pair['key'],
-        algorithm=self._default_digest_class(),
-        backend=_cryptography_backend,
-      )
-      crl_pem = crl.public_bytes(serialization.Encoding.PEM)
-      self._storage.storeCertificateRevocationList(
-        crl_pem,
-        expiration_date=utils.datetime2timestamp(now + self._crl_renew_time),
-      )
-    return crl_pem
+          ],
+          revoked_certificates=revoked_certificate_list,
+        ).sign(
+          private_key=ca_key_pair['key'],
+          algorithm=self._default_digest_class(),
+          backend=_cryptography_backend,
+        ).public_bytes(serialization.Encoding.PEM)
+        crl_pem_dict[authority_key_identifier] = (
+          crl_pem,
+          last_update + self._crl_renew_time,
+        )
+      result[authority_key_identifier] = crl_pem
+    return result
 
 class UserCertificateAuthority(CertificateAuthority):
   """
@@ -791,10 +848,10 @@ class UserCertificateAuthority(CertificateAuthority):
     certificates.
     """
     ca_cert_list = self.getCACertificateList()
-    crl = x509.load_pem_x509_crl(
-      self.getCertificateRevocationList(),
-      _cryptography_backend,
-    )
+    crl_list = [
+      x509.load_pem_x509_crl(x, _cryptography_backend)
+      for x in self.getCertificateRevocationListDict().itervalues()
+    ]
     signing_key = os.urandom(32)
     symetric_key = os.urandom(32)
     iv = os.urandom(16)
@@ -813,7 +870,7 @@ class UserCertificateAuthority(CertificateAuthority):
     key_list = []
     for crt_pem in self._storage.iterCertificates():
       try:
-        crt = utils.load_certificate(crt_pem, ca_cert_list, crl)
+        crt = utils.load_certificate(crt_pem, ca_cert_list, crl_list)
       except CertificateVerificationError:
         continue
       public_key = crt.public_key()
