@@ -113,6 +113,18 @@ class CLICaucaseClient(object):
     kw.setdefault('file', self._stdout)
     print(*args, **kw)
 
+  def __enter__(self):
+    return self
+
+  def __exit__(self, exc_type, exc_value, traceback):
+    self.close()
+
+  def close(self):
+    """
+    Tell client to close any open connection.
+    """
+    self._client.close()
+
   def putCSR(self, csr_path_list):
     """
     --send-csr
@@ -403,13 +415,27 @@ def main(argv=None, stdout=sys.stdout, stderr=sys.stderr):
     '--crl',
     default='cas.crl.pem',
     metavar='CRL_PATH',
-    help='Services certificate revocation list location. default: %(default)s',
+    help='Services certificate revocation list location. '
+    'May be an existing directory or file, or non-existing. '
+    'If non-existing and given path has an extension, a file will be created, '
+    'otherwise a directory will be. '
+    'When it is a file, it may contain multiple PEM-encoded concatenated '
+    'CRLs. When it is a directory, it may contain multiple files, each '
+    'containing a single PEM-encoded CRL. '
+    'default: %(default)s',
   )
   parser.add_argument(
     '--user-crl',
     default='cau.crl.pem',
     metavar='CRL_PATH',
-    help='Users certificate revocation list location. default: %(default)s',
+    help='Users certificate revocation list location. '
+    'May be an existing directory or file, or non-existing. '
+    'If non-existing and given path has an extension, a file will be created, '
+    'otherwise a directory will be. '
+    'When it is a file, it may contain multiple PEM-encoded concatenated '
+    'CRLs. When it is a directory, it may contain multiple files, each '
+    'containing a single PEM-encoded CRL. '
+    'default: %(default)s',
   )
   parser.add_argument(
     '--threshold',
@@ -596,7 +622,7 @@ def main(argv=None, stdout=sys.stdout, stderr=sys.stderr):
     if args.update_user or args.mode == MODE_USER:
       updated |= CaucaseClient.updateCAFile(cau_url, args.user_ca_crt)
 
-    client = CLICaucaseClient(
+    with CLICaucaseClient(
       client=CaucaseClient(
         ca_url={
           MODE_SERVICE: cas_url,
@@ -607,37 +633,37 @@ def main(argv=None, stdout=sys.stdout, stderr=sys.stderr):
       ),
       stdout=stdout,
       stderr=stderr,
-    )
-    ca_list = [
-      utils.load_ca_certificate(x)
-      for x in utils.getCertList({
-        MODE_SERVICE: args.ca_crt,
-        MODE_USER: args.user_ca_crt,
-      }[args.mode])
-    ]
-    client.putCSR(args.send_csr)
-    client.getCSR(args.get_csr)
-    warning, error = client.getCRT(warning, error, args.get_crt, ca_list)
-    error = client.revokeCRT(error, args.revoke_crt)
-    updated, error = client.renewCRT(
-      crt_key_list=args.renew_crt,
-      renewal_deadline=datetime.datetime.utcnow() + datetime.timedelta(
-        args.threshold,
-        0,
-      ),
-      key_len=args.key_len,
-      ca_certificate_list=ca_list,
-      updated=updated,
-      error=error,
-    )
-    client.signCSR(args.sign_csr)
-    client.signCSRWith(args.sign_csr_with)
-    client.rejectCSR(args.reject_csr)
-    client.revokeOtherCRT(args.revoke_other_crt)
-    client.revokeSerial(args.revoke_serial)
-    # show latest CSR list status
-    if args.list_csr:
-      client.listCSR(args.mode)
+    ) as client:
+      ca_list = [
+        utils.load_ca_certificate(x)
+        for x in utils.getCertList({
+          MODE_SERVICE: args.ca_crt,
+          MODE_USER: args.user_ca_crt,
+        }[args.mode])
+      ]
+      client.putCSR(args.send_csr)
+      client.getCSR(args.get_csr)
+      warning, error = client.getCRT(warning, error, args.get_crt, ca_list)
+      error = client.revokeCRT(error, args.revoke_crt)
+      updated, error = client.renewCRT(
+        crt_key_list=args.renew_crt,
+        renewal_deadline=datetime.datetime.utcnow() + datetime.timedelta(
+          args.threshold,
+          0,
+        ),
+        key_len=args.key_len,
+        ca_certificate_list=ca_list,
+        updated=updated,
+        error=error,
+      )
+      client.signCSR(args.sign_csr)
+      client.signCSRWith(args.sign_csr_with)
+      client.rejectCSR(args.reject_csr)
+      client.revokeOtherCRT(args.revoke_other_crt)
+      client.revokeSerial(args.revoke_serial)
+      # show latest CSR list status
+      if args.list_csr:
+        client.listCSR(args.mode)
     # update our CRL after all revocations we were requested
     updated |= CaucaseClient.updateCRLFile(cas_url, args.crl, [
       utils.load_ca_certificate(x)
@@ -681,17 +707,16 @@ def probe(argv=None):
   )
   ca_url, = parser.parse_args(argv).ca_url
   cas_url = ca_url + '/cas'
-  http_client = CaucaseClient(
-    ca_url=cas_url,
-  )
-  http_ca_pem = http_client.getCACertificate()
-  https_ca_pem = HTTPSOnlyCaucaseClient(
-    ca_url=cas_url,
-    ca_crt_pem_list=[http_ca_pem],
-  ).getCACertificate()
-  # Retrieve again in case there was a renewal between both calls - we do
-  # not expect 2 renewals in very short succession.
-  http2_ca_pem = http_client.getCACertificate()
+  with CaucaseClient(ca_url=cas_url) as http_client:
+    http_ca_pem = http_client.getCACertificate()
+    with HTTPSOnlyCaucaseClient(
+      ca_url=cas_url,
+      ca_crt_pem_list=[http_ca_pem],
+    ) as https_client:
+      https_ca_pem = https_client.getCACertificate()
+    # Retrieve again in case there was a renewal between both calls - we do
+    # not expect 2 renewals in very short succession.
+    http2_ca_pem = http_client.getCACertificate()
   if https_ca_pem not in (http_ca_pem, http2_ca_pem):
     raise ValueError('http and https do not serve the same caucase database')
 
@@ -814,9 +839,16 @@ def updater(argv=None, until=utils.until):
     required=True,
     metavar='CRT_PATH',
     help='Path of your certificate revocation list for MODE. '
+    'May be an existing directory or file, or non-existing. '
+    'If non-existing and given path has an extension, a file will be created, '
+    'otherwise a directory will be. '
+    'When it is a file, it may contain multiple PEM-encoded concatenated '
+    'CRLs. When it is a directory, it may contain multiple files, each '
+    'containing a single PEM-encoded CRL. '
     'Will be maintained up-to-date.'
   )
   args = parser.parse_args(argv)
+  client = None
   try:
     cas_url = args.ca_url + '/cas'
     ca_url = {
@@ -872,6 +904,7 @@ def updater(argv=None, until=utils.until):
         cas_url,
         args.cas_ca,
       ):
+        client.close()
         client = RetryingCaucaseClient(
           ca_url=ca_url,
           ca_crt_pem_list=utils.getCertList(args.cas_ca)
@@ -889,11 +922,11 @@ def updater(argv=None, until=utils.until):
       if RetryingCaucaseClient.updateCRLFile(ca_url, args.crl, ca_crt_list):
         print('Got new CRL')
         updated = True
-      with open(args.crl, 'rb') as crl_file:
+      for crl_pem in utils.getCRLList(args.crl):
         next_deadline = min(
           next_deadline,
           utils.load_crl(
-            crl_file.read(),
+            crl_pem,
             ca_crt_list,
           ).next_update - crl_threshold,
         )
@@ -942,6 +975,9 @@ def updater(argv=None, until=utils.until):
     # Not intercepting KeyboardInterrupt so interrupting outside of
     # interruptibleSleep shows where the script got interrupted.
     pass
+  finally:
+    if client is not None:
+      client.close()
 
 def rerequest(argv=None):
   """
