@@ -308,7 +308,10 @@ def getSSLContext(
   # certificate.
   #ssl_context.verify_flags = ssl.VERIFY_CRL_CHECK_LEAF
   ssl_context.load_verify_locations(
-    cadata=utils.toUnicode(cau.getCACertificate()),
+    cadata=utils.toUnicode(b'\n'.join(
+      utils.dump_certificate(x)
+      for x in cau.getCACertificateList()
+    )),
   )
   http_cas_certificate_list = http_cas.getCACertificateList()
   threshold_delta = datetime.timedelta(threshold, 0)
@@ -363,7 +366,6 @@ def getSSLContext(
           extensions=[
             Extension(
               x509.KeyUsage(
-                # pylint: disable=bad-whitespace
                 digital_signature =True,
                 content_commitment=False,
                 key_encipherment  =True,
@@ -373,7 +375,6 @@ def getSSLContext(
                 crl_sign          =False,
                 encipher_only     =False,
                 decipher_only     =False,
-                # pylint: enable=bad-whitespace
               ),
               critical=True,
             ),
@@ -808,7 +809,7 @@ def main(
     backup_period = datetime.timedelta(args.backup_period, 0)
     try:
       next_backup = max(
-        datetime.datetime.utcfromtimestamp(os.stat(x).st_ctime)
+        utils.timestamp2datetime(os.stat(x).st_ctime)
         for x in glob.iglob(
           os.path.join(args.backup_directory, '*' + BACKUP_SUFFIX),
         )
@@ -1012,11 +1013,12 @@ def manage(argv=None, stdout=sys.stdout):
     import_ca_dict = defaultdict(
       (lambda: {'crt': None, 'key': None, 'from': []}),
     )
-    for import_ca in args.import_ca:
-      with open(import_ca, 'rb') as ca_file:
+    password = b''
+    for import_ca_path in args.import_ca:
+      with open(import_ca_path, 'rb') as ca_file:
         ca_data = ca_file.read()
       for index, component in enumerate(pem.parse(ca_data)):
-        name = '%r, block %i' % (import_ca, index)
+        name = '%r, block %i' % (import_ca_path, index)
         if isinstance(component, pem.Certificate):
           component_name = 'crt'
           component_value = x509.load_pem_x509_certificate(
@@ -1024,7 +1026,6 @@ def manage(argv=None, stdout=sys.stdout):
             _cryptography_backend,
           )
         elif isinstance(component, pem.Key):
-          password = None
           while True:
             component_name = 'key'
             try:
@@ -1113,17 +1114,27 @@ def manage(argv=None, stdout=sys.stdout):
     db = SQLite3Storage(db_path, table_prefix='cas')
     trusted_ca_crt_set = [
       utils.load_ca_certificate(x['crt_pem'])
-      for x in db.getCAKeyPairList()
+      for x in db.getCAKeyPairList(prune=False)
     ]
     latest_ca_not_after = max(
       x.not_valid_after
       for x in trusted_ca_crt_set
     )
     already_revoked_count = revoked_count = 0
+    crl_number = crl_last_update = None
     for import_crl in args.import_crl:
       with open(import_crl, 'rb') as crl_file:
-        crl_data = crl_file.read()
-      for revoked in utils.load_crl(crl_data, trusted_ca_crt_set):
+        crl = utils.load_crl(crl_file.read(), trusted_ca_crt_set)
+      current_crl_number = crl.extensions.get_extension_for_class(
+        x509.CRLNumber,
+      ).value.crl_number
+      if crl_number is None:
+        crl_number = current_crl_number
+        crl_last_update = crl.last_update
+      else:
+        crl_number = max(crl_number, current_crl_number)
+        crl_last_update = max(crl_last_update, crl.last_update)
+      for revoked in crl:
         try:
           db.revoke(
             revoked.serial_number,
@@ -1133,6 +1144,15 @@ def manage(argv=None, stdout=sys.stdout):
           already_revoked_count += 1
         else:
           revoked_count += 1
+    db.storeCRLLastUpdate(utils.datetime2timestamp(crl_last_update))
+    db.storeCRLNumber(crl_number)
+    print(
+      'Set CRL number to %i and last update to %s' % (
+        crl_number,
+        crl_last_update.isoformat(' '),
+      ),
+      file=stdout,
+    )
     print(
       'Revoked %i certificates (%i were already revoked)' % (
         revoked_count,
@@ -1141,15 +1161,24 @@ def manage(argv=None, stdout=sys.stdout):
       file=stdout,
     )
   if args.export_ca is not None:
-    encryption_algorithm = serialization.BestAvailableEncryption(
-      getBytePass('CA export passphrase: ')
-    )
+    if os.path.exists(args.export_ca):
+      parser.error('file exists: %r' % (args.export_ca, ))
+    while True:
+      passphrase = getBytePass(
+        'CA export passphrase: ',
+      )
+      if getBytePass(
+        '             (again): ',
+      ) == passphrase:
+        break
+      print('Error: Input does not match, retrying')
+    encryption_algorithm = serialization.BestAvailableEncryption(passphrase)
     with open(args.export_ca, 'wb') as export_ca_file:
       write = export_ca_file.write
       for key_pair in SQLite3Storage(
         db_path,
         table_prefix='cas',
-      ).getCAKeyPairList():
+      ).getCAKeyPairList(prune=False):
         write(
           key_pair['crt_pem'] + serialization.load_pem_private_key(
             key_pair['key_pem'],
