@@ -27,11 +27,14 @@ import os
 import sqlite3
 from threading import local
 from time import time
+from cryptography import x509
+from cryptography.hazmat.backends import default_backend
 from .exceptions import NoStorage, NotFound, Found
-from .utils import toBytes, toUnicode
+from .utils import toBytes, toUnicode, datetime2timestamp
 
 __all__ = ('SQLite3Storage', )
 
+_cryptography_backend = default_backend()
 DAY_IN_SECONDS = 60 * 60 * 24
 
 class NoReentryConnection(sqlite3.Connection):
@@ -44,7 +47,7 @@ class NoReentryConnection(sqlite3.Connection):
   __entered = False
 
   def __enter__(self):
-    if self.__entered:
+    if self.__entered: # pragma: no cover
       raise Exception('Subtransactions are not supported')
     self.__entered = True
     return super(NoReentryConnection, self).__enter__()
@@ -113,7 +116,8 @@ class SQLite3Storage(local):
       # sqlite can accept as integers, so store these as text. Use a trivial
       # string serialisation: not very space efficient, but this should not be
       # a limiting issue for our use-cases anyway.
-      db.cursor().executescript('''
+      c = db.cursor()
+      c.executescript('''
         CREATE TABLE IF NOT EXISTS %(prefix)sca (
           expiration_date INTEGER,
           key TEXT,
@@ -131,10 +135,6 @@ class SQLite3Storage(local):
           revocation_date INTEGER,
           expiration_date INTEGER
         );
-        CREATE TABLE IF NOT EXISTS %(prefix)scrl (
-          expiration_date INTEGER,
-          crl TEXT
-        );
         CREATE TABLE IF NOT EXISTS %(prefix)scounter (
           name TEXT PRIMARY KEY,
           value INTEGER
@@ -142,11 +142,42 @@ class SQLite3Storage(local):
         CREATE TABLE IF NOT EXISTS %(prefix)sconfig_once (
           name TEXT PRIMARY KEY,
           value TEXT
-        )
+        );
+        CREATE TABLE IF NOT EXISTS %(prefix)sconfig (
+          name TEXT PRIMARY KEY,
+          value TEXT
+        );
       ''' % {
         'prefix': table_prefix,
         'key_id_constraint': 'UNIQUE' if enforce_unique_key_id else '',
       })
+      try:
+        crl_row = self._executeSingleRow(
+          'SELECT crl FROM %(prefix)scrl '
+          'ORDER BY expiration_date DESC LIMIT 1',
+        )
+      except sqlite3.OperationalError as exc:
+        # XXX: no error codes in sqlite and only a generic error class ?
+        if not exc.args[0].startswith('no such table: '): # pragma: no cover
+          raise
+        crl_row = None
+      if crl_row is not None:
+        self._setConfig(
+          'crl_last_update',
+          str(datetime2timestamp(x509.load_pem_x509_crl(
+            toBytes(crl_row['crl']),
+            _cryptography_backend,
+          ).last_update)),
+        )
+        self._execute(c, 'DROP TABLE IF EXISTS %(prefix)scrl')
+
+  def _execute(self, cursor, sql, parameters=()):
+    return cursor.execute(
+      sql % {
+        'prefix': self._table_prefix,
+      },
+      parameters,
+    )
 
   def _incrementCounter(self, name, increment=1, initial=0):
     """
@@ -155,9 +186,7 @@ class SQLite3Storage(local):
     Does not commit.
     """
     row = self._executeSingleRow(
-      'SELECT value FROM %scounter WHERE name = ? LIMIT 2' % (
-        self._table_prefix,
-      ),
+      'SELECT value FROM %(prefix)scounter WHERE name = ? LIMIT 2',
       (name, ),
     )
     if row is None:
@@ -165,10 +194,9 @@ class SQLite3Storage(local):
     else:
       value = row['value']
     value += increment
-    self._db.cursor().execute(
-      'INSERT OR REPLACE INTO %scounter (name, value) VALUES (?, ?)' % (
-        self._table_prefix,
-      ),
+    self._execute(
+      self._db.cursor(),
+      'INSERT OR REPLACE INTO %(prefix)scounter (name, value) VALUES (?, ?)',
       (name, value),
     )
     return value
@@ -177,7 +205,7 @@ class SQLite3Storage(local):
     """
     Execute <sql>, raise if it produces more than 1 row, and return it.
     """
-    result_list = self._db.cursor().execute(sql, parameters).fetchall()
+    result_list = self._execute(self._db.cursor(), sql, parameters).fetchall()
     if result_list:
       result, = result_list
       return result
@@ -190,9 +218,7 @@ class SQLite3Storage(local):
     """
     with self._db:
       result = self._executeSingleRow(
-        'SELECT value FROM %sconfig_once WHERE name = ?' % (
-          self._table_prefix,
-        ),
+        'SELECT value FROM %(prefix)sconfig_once WHERE name = ?',
         (name, ),
       )
     if result is None:
@@ -206,37 +232,59 @@ class SQLite3Storage(local):
     """
     try:
       with self._db as db:
-        db.cursor().execute(
-          'INSERT INTO %sconfig_once (name, value) VALUES (?, ?)' % (
-            self._table_prefix,
-          ),
+        self._execute(
+          db.cursor(),
+          'INSERT INTO %(prefix)sconfig_once (name, value) VALUES (?, ?)',
           (name, value),
         )
     except sqlite3.IntegrityError:
       pass
 
-  def getCAKeyPairList(self):
+  def _getConfig(self, name, default):
+    """
+    Retrieve the value of <name> from config list, or <default> if not
+    stored.
+    """
+    result = self._executeSingleRow(
+      'SELECT value FROM %(prefix)sconfig WHERE name = ?',
+      (name, ),
+    )
+    if result is None:
+      return default
+    return result['value']
+
+  def _setConfig(self, name, value):
+    """
+    Store <value> as <name> in config list, possibly overwriting an existing
+    entry.
+    """
+    self._execute(
+      self._db.cursor(),
+      'INSERT OR REPLACE INTO %(prefix)sconfig (name, value) VALUES (?, ?)',
+      (name, value),
+    )
+
+  def getCAKeyPairList(self, prune=True):
     """
     Return the chronologically sorted (oldest in [0], newest in [-1])
     certificate authority key pairs.
     """
     with self._db as db:
       c = db.cursor()
-      c.execute(
-        'DELETE FROM %sca WHERE expiration_date < ?' % (
-          self._table_prefix,
-        ),
-        (time(), ),
-      )
+      if prune:
+        self._execute(
+          c,
+          'DELETE FROM %(prefix)sca WHERE expiration_date < ?',
+          (time(), ),
+        )
       return [
         {
           'crt_pem': toBytes(x['crt']),
           'key_pem': toBytes(x['key']),
         }
-        for x in db.cursor().execute(
-          'SELECT key, crt FROM %sca ORDER BY expiration_date ASC' % (
-            self._table_prefix,
-          ),
+        for x in self._execute(
+          c,
+          'SELECT key, crt FROM %(prefix)sca ORDER BY expiration_date ASC',
         ).fetchall()
       ]
 
@@ -245,14 +293,14 @@ class SQLite3Storage(local):
     Store a certificate authority key pair.
     expiration_timestamp (int)
       Unix GMT timestamp of CA certificate "valid until" date.
-    key_pair (dict with 'key' and 'crt' items)
+    key_pair (dict with 'key_pem' and 'crt_pem' items)
       CA key pair to store, as bytes.
     """
     with self._db as db:
-      db.cursor().execute(
-        'INSERT INTO %sca (expiration_date, key, crt) VALUES (?, ?, ?)' % (
-          self._table_prefix,
-        ),
+      self._execute(
+        db.cursor(),
+        'INSERT INTO %(prefix)sca '
+        '(expiration_date, key, crt) VALUES (?, ?, ?)',
         (
           expiration_timestamp,
           key_pair['key_pem'],
@@ -273,9 +321,7 @@ class SQLite3Storage(local):
     """
     with self._db as db:
       known_csr = self._executeSingleRow(
-        'SELECT id FROM %scrt WHERE csr = ? LIMIT 2' % (
-          self._table_prefix,
-        ),
+        'SELECT id FROM %(prefix)scrt WHERE csr = ? LIMIT 2',
         (csr_pem, ),
       )
       if known_csr is not None:
@@ -287,28 +333,24 @@ class SQLite3Storage(local):
         requested_count = None
       else:
         if self._executeSingleRow(
-          'SELECT COUNT(*) FROM %scrt WHERE crt IS NULL' % (
-            self._table_prefix,
-          )
+          'SELECT COUNT(*) FROM %(prefix)scrt WHERE crt IS NULL',
         )[0] >= self._max_csr_amount:
           raise NoStorage
         requested_count = self._incrementCounter('received_csr')
       csr_id = getrandbits(63)
       c = db.cursor()
-      c.execute(
-        'INSERT INTO %scrt (id, key_id, csr) VALUES (?, ?, ?)' % (
-          self._table_prefix,
-        ),
+      self._execute(
+        c,
+        'INSERT INTO %(prefix)scrt (id, key_id, csr) VALUES (?, ?, ?)',
         (
           csr_id,
           key_id,
           csr_pem,
         ),
       )
-      c.execute(
-        'DELETE FROM %scrt WHERE expiration_date < ?' % (
-          self._table_prefix,
-        ),
+      self._execute(
+        c,
+        'DELETE FROM %(prefix)scrt WHERE expiration_date < ?',
         (time(), ),
       )
       return csr_id, requested_count
@@ -321,10 +363,9 @@ class SQLite3Storage(local):
     """
     with self._db as db:
       c = db.cursor()
-      c.execute(
-        'DELETE FROM %scrt WHERE id = ? AND crt IS NULL' % (
-          self._table_prefix,
-        ),
+      self._execute(
+        c,
+        'DELETE FROM %(prefix)scrt WHERE id = ? AND crt IS NULL',
         (csr_id, ),
       )
       if c.rowcount == 1:
@@ -342,9 +383,7 @@ class SQLite3Storage(local):
     """
     with self._db:
       result = self._executeSingleRow(
-        'SELECT csr FROM %scrt WHERE id = ?' % (
-          self._table_prefix,
-        ),
+        'SELECT csr FROM %(prefix)scrt WHERE id = ?',
         (csr_id, ),
       )
       if result is None:
@@ -367,10 +406,9 @@ class SQLite3Storage(local):
           # to then have to unicode-ify, just unicode-ify here.
           'csr': toUnicode(x['csr']),
         }
-        for x in db.cursor().execute(
-          'SELECT id, csr FROM %scrt WHERE crt IS NULL' % (
-            self._table_prefix,
-          ),
+        for x in self._execute(
+          db.cursor(),
+          'SELECT id, csr FROM %(prefix)scrt WHERE crt IS NULL',
         ).fetchall()
       ]
 
@@ -383,11 +421,10 @@ class SQLite3Storage(local):
     """
     with self._db as db:
       c = db.cursor()
-      c.execute(
-        'UPDATE %scrt SET crt=?, expiration_date = ? '
-        'WHERE id = ? AND crt IS NULL' % (
-          self._table_prefix,
-        ),
+      self._execute(
+        c,
+        'UPDATE %(prefix)scrt SET crt=?, expiration_date = ? '
+        'WHERE id = ? AND crt IS NULL',
         (
           crt,
           int(time() + self._crt_keep_time),
@@ -409,20 +446,17 @@ class SQLite3Storage(local):
     """
     with self._db as db:
       row = self._executeSingleRow(
-        'SELECT crt, expiration_date FROM %scrt '
-        'WHERE id = ? AND crt IS NOT NULL' % (
-          self._table_prefix,
-        ),
+        'SELECT crt, expiration_date FROM %(prefix)scrt '
+        'WHERE id = ? AND crt IS NOT NULL',
         (crt_id, ),
       )
       if row is None:
         raise NotFound
       new_expiration_date = int(time() + self._crt_read_keep_time)
       if row['expiration_date'] > new_expiration_date:
-        db.cursor().execute(
-          'UPDATE %scrt SET expiration_date = ? WHERE id = ?' % (
-            self._table_prefix,
-          ),
+        self._execute(
+          db.cursor(),
+          'UPDATE %(prefix)scrt SET expiration_date = ? WHERE id = ?',
           (
             new_expiration_date,
             crt_id,
@@ -439,9 +473,7 @@ class SQLite3Storage(local):
     """
     with self._db:
       row = self._executeSingleRow(
-        'SELECT crt FROM %scrt WHERE key_id = ? AND crt IS NOT NULL' % (
-          self._table_prefix,
-        ),
+        'SELECT crt FROM %(prefix)scrt WHERE key_id = ? AND crt IS NOT NULL',
         (key_id, ),
       )
       if row is None:
@@ -454,9 +486,10 @@ class SQLite3Storage(local):
     """
     with self._db as db:
       c = db.cursor()
-      c.execute('SELECT crt FROM %scrt WHERE crt IS NOT NULL' % (
-        self._table_prefix,
-      ))
+      self._execute(
+        c,
+        'SELECT crt FROM %(prefix)scrt WHERE crt IS NOT NULL',
+      )
       while True:
         row = c.fetchone()
         if row is None:
@@ -476,65 +509,65 @@ class SQLite3Storage(local):
     """
     with self._db as db:
       c = db.cursor()
-      c.execute('DELETE FROM %scrl' % (
-        self._table_prefix,
-      ))
       try:
-        c.execute(
-          'INSERT INTO %srevoked '
+        self._execute(
+          c,
+          'INSERT INTO %(prefix)srevoked '
           '(serial, revocation_date, expiration_date) '
-          'VALUES (?, ?, ?)' % (
-            self._table_prefix,
-          ),
+          'VALUES (?, ?, ?)',
           (
             str(serial),
             int(time()),
             expiration_date,
-          )
+          ),
         )
       except sqlite3.IntegrityError:
         raise Found
-
-  def getCertificateRevocationList(self):
-    """
-    Get PEM-encoded current Certificate Revocation List.
-
-    Returns None if there is no CRL.
-    """
-    with self._db:
-      row = self._executeSingleRow(
-        'SELECT crl FROM %scrl '
-        'WHERE expiration_date > ? ORDER BY expiration_date DESC LIMIT 1' % (
-          self._table_prefix,
-        ),
-        (time(), )
-      )
-      if row is not None:
-        return toBytes(row['crl'])
-    return None
+      self._incrementCounter('crl_number')
 
   def getNextCertificateRevocationListNumber(self):
     """
     Get next CRL sequence number.
     """
-    return self._incrementCounter('crl_number')
+    with self._db:
+      return self._incrementCounter('crl_number')
 
-  def storeCertificateRevocationList(self, crl, expiration_date):
+  def storeCRLLastUpdate(self, last_update):
     """
-    Store Certificate Revocation List.
+    Get next CRL sequence number.
+    """
+    with self._db:
+      self._setConfig('crl_last_update', str(last_update))
+
+  def storeCRLNumber(self, crl_number):
+    """
+    Set the current CRL sequence number.
+    Use only when importing an existing CA.
     """
     with self._db as db:
-      c = db.cursor()
-      c.execute('DELETE FROM %scrl' % (
-        self._table_prefix,
-      ))
-      c.execute(
-        'INSERT INTO %scrl (expiration_date, crl) VALUES (?, ?)' % (
-          self._table_prefix,
-        ),
+      self._execute(
+        db.cursor(),
+        'INSERT OR REPLACE INTO %(prefix)scounter (name, value) VALUES (?, ?)',
         (
-          int(expiration_date),
-          crl,
+          'crl_number',
+          crl_number,
+        ),
+      )
+
+  def getCurrentCRLNumberAndLastUpdate(self):
+    """
+    Get the current CRL sequence number.
+    """
+    with self._db:
+      last_update = self._getConfig('crl_last_update', None)
+      return (
+        # Note: does not increment the counter, but may set it to the default
+        # value.
+        self._incrementCounter('crl_number', increment=0),
+        (
+          last_update
+          if last_update is None else
+          int(last_update, 10)
         ),
       )
 
@@ -550,10 +583,9 @@ class SQLite3Storage(local):
     """
     with self._db as db:
       c = db.cursor()
-      c.execute(
-        'DELETE FROM %srevoked WHERE expiration_date < ?' % (
-          self._table_prefix,
-        ),
+      self._execute(
+        c,
+        'DELETE FROM %(prefix)srevoked WHERE expiration_date < ?',
         (time(), ),
       )
       return [
@@ -561,10 +593,9 @@ class SQLite3Storage(local):
           'revocation_date': int(x['revocation_date']),
           'serial': int(x['serial']),
         }
-        for x in c.execute(
-          'SELECT revocation_date, serial FROM %srevoked' % (
-            self._table_prefix,
-          ),
+        for x in self._execute(
+          c,
+          'SELECT revocation_date, serial FROM %(prefix)srevoked',
         )
       ]
 
