@@ -24,7 +24,7 @@ Caucase - Certificate Authority for Users, Certificate Authority for SErvices
 Small-ish functions needed in many places.
 """
 from __future__ import absolute_import, print_function
-from binascii import a2b_base64, b2a_base64
+from binascii import a2b_base64, b2a_base64, hexlify
 import calendar
 import codecs
 from collections import defaultdict
@@ -32,6 +32,7 @@ import datetime
 import email
 import json
 import os
+import sys
 import threading
 import traceback
 import time
@@ -86,7 +87,6 @@ _CAUCASE_LEGACY_OID_AUTO_SIGNED = x509.oid.ObjectIdentifier(
   CAUCASE_LEGACY_OID_AUTO_SIGNED,
 )
 
-
 def isCertificateAutoSigned(crt):
   """
   Checks whether given certificate was automatically signed by caucase.
@@ -124,26 +124,31 @@ def _getPEMTypeDict(path, result=None):
 def getCertList(crt_path):
   """
   Return a list of certificates.
-  Raises if there is anything else than a certificate.
   """
-  if not os.path.exists(crt_path):
+  return _getPEMListFromPath(crt_path, pem.Certificate)
+
+def getCRLList(crl_path):
+  """
+  Return a list of Certificate Revocation Lists.
+  """
+  return _getPEMListFromPath(crl_path, pem.CertificateRevocationList)
+
+def _getPEMListFromPath(path, pem_type):
+  if not os.path.exists(path):
     return []
-  if os.path.isdir(crt_path):
-    file_list = [os.path.join(crt_path, x) for x in os.listdir(crt_path)]
-  else:
-    file_list = [crt_path]
-  result = []
-  for file_name in file_list:
-    type_dict = _getPEMTypeDict(file_name)
-    crt_list = type_dict.pop(pem.Certificate)
-    if type_dict:
-      raise ValueError('%s contains more than just certificates' % (file_name, ))
-    result.extend(x.as_bytes() for x in crt_list)
-  return result
+  return [
+    pem_object.as_bytes()
+    for file_name in (
+      [os.path.join(path, x) for x in os.listdir(path)]
+      if os.path.isdir(path) else
+      [path]
+    )
+    for pem_object in _getPEMTypeDict(file_name).get(pem_type, ())
+  ]
 
 def saveCertList(crt_path, cert_pem_list):
   """
-  Store given list of PEm-encoded certificates in given path.
+  Store given list of PEM-encoded certificates in given path.
 
   crt_path (str)
   May point to a directory a file, or nothing.
@@ -154,61 +159,90 @@ def saveCertList(crt_path, cert_pem_list):
 
   cert_pem_list (list of bytes)
   """
-  if os.path.exists(crt_path):
-    if os.path.isfile(crt_path):
-      saveCertListTo = _saveCertListToFile
-    elif os.path.isdir(crt_path):
-      saveCertListTo = _saveCertListToDirectory
+  _savePEMList(crt_path, cert_pem_list, load_ca_certificate, '.ca.pem')
+
+def saveCRLList(crl_path, crl_pem_list):
+  """
+  Store given list of PEM-encoded Certificate Revocation Lists in given path.
+
+  crl_path (str)
+  May point to a directory a file, or nothing.
+  If it does not exist, and this value contains an extension, a file is
+  created, otherwise a directory is.
+  If it is a file, all CRLs are written in it.
+  If it is a folder, each CRL is stored in a separate file.
+
+  crl_pem_list (list of bytes)
+  """
+  _savePEMList(
+    crl_path,
+    crl_pem_list,
+    lambda x: x509.load_pem_x509_crl(x, _cryptography_backend),
+    '.crl.pem',
+  )
+
+def _savePEMList(path, pem_list, pem_loader, extension):
+  if os.path.exists(path):
+    if os.path.isfile(path):
+      savePEMList = _savePEMListToFile
+    elif os.path.isdir(path):
+      savePEMList = _savePEMListToDirectory
     else:
       raise TypeError('%s exist and is neither a directory nor a file' % (
-        crt_path,
+        path,
       ))
   else:
-    saveCertListTo = (
-      _saveCertListToFile
-      if os.path.splitext(crt_path)[1] else
-      _saveCertListToDirectory
+    savePEMList = (
+      _savePEMListToFile
+      if os.path.splitext(path)[1] else
+      _savePEMListToDirectory
     )
-  saveCertListTo(crt_path, cert_pem_list)
+  savePEMList(path, pem_list, pem_loader, extension)
 
-def _saveCertListToFile(ca_crt_path, cert_pem_list):
-  with open(ca_crt_path, 'wb') as ca_crt_file:
-    ca_crt_file.write(b''.join(cert_pem_list))
+def _savePEMListToFile(file_path, pem_list, pem_loader, extension):
+  _ = pem_loader # Silence pylint
+  _ = extension # Silence pylint
+  with open(file_path, 'wb') as pem_file:
+    for pem_chunk in pem_list:
+      pem_file.write(pem_chunk)
 
-def _saveCertListToDirectory(crt_dir, cert_pem_list):
-  if not os.path.exists(crt_dir):
-    os.mkdir(crt_dir)
-  ca_cert_dict = {
-    '%x.pem' % (load_ca_certificate(x).serial_number, ): x
-    for x in cert_pem_list
+def _savePEMListToDirectory(dir_path, pem_list, pem_loader, extension):
+  if not os.path.exists(dir_path):
+    os.mkdir(dir_path)
+  pem_dict = {
+    hexlify(
+      pem_loader(x).extensions.get_extension_for_class(
+        x509.AuthorityKeyIdentifier,
+      ).value.key_identifier,
+    ).decode('ascii') + extension: x
+    for x in pem_list
   }
-  for cert_filename in os.listdir(crt_dir):
-    ca_crt_path = os.path.join(crt_dir, cert_filename)
-    if not os.path.isfile(ca_crt_path):
-      # Not a file and not a symlink to a file, ignore
+  for filename in os.listdir(dir_path):
+    filepath = os.path.join(dir_path, filename)
+    if not filepath.endswith(extension) or not os.path.isfile(filepath):
+      # Not a managed file name and not a symlink to a file, ignore
       continue
-    if not os.path.islink(ca_crt_path) and cert_filename in ca_cert_dict:
+    if not os.path.islink(filepath) and filename in pem_dict:
       try:
         # pylint: disable=unbalanced-tuple-unpacking
-        cert, = getCertList(ca_crt_path)
+        file_pem_item, = _getPEMTypeDict(filepath).itervalues()
         # pylint: enable=unbalanced-tuple-unpacking
       # pylint: disable=broad-except
       except Exception:
       # pylint: enable=broad-except
-        # Inconsistent content (multiple certificates, not CA certificates,
-        # ...): overwrite file
+        # File contains multiple PEM items: overwrite
         pass
       else:
-        if cert == ca_cert_dict[cert_filename]:
+        if file_pem_item == pem_dict[filename]:
           # Already consistent, do not edit.
-          del ca_cert_dict[cert_filename]
+          del pem_dict[filename]
     else:
       # Unknown file (ex: expired certificate), or a symlink to a file: delete
-      os.unlink(ca_crt_path)
-  for cert_filename, cert_pem in ca_cert_dict.items():
-    ca_crt_path = os.path.join(crt_dir, cert_filename)
-    with open(ca_crt_path, 'wb') as ca_crt_file:
-      ca_crt_file.write(cert_pem)
+      os.unlink(filepath)
+  for filename, pem_item in pem_dict.iteritems():
+    filepath = os.path.join(dir_path, filename)
+    with open(filepath, 'wb') as pem_file:
+      pem_file.write(pem_item)
 
 def getCert(crt_path):
   """
@@ -219,7 +253,7 @@ def getCert(crt_path):
   crt, = type_dict.get(pem.Certificate)
   return crt.as_bytes()
 
-def getCertKeyAndCACert(crt_path, crl):
+def getCertKeyAndCACert(crt_path, crl_list):
   """
   Return a certificate with its private key and the certificate which signed
   it.
@@ -241,7 +275,7 @@ def getCertKeyAndCACert(crt_path, crl):
     except ValueError:
       continue
     # key and crt match, check signatures
-    load_certificate(crt, [load_ca_certificate(ca_crt)], crl)
+    load_certificate(crt, [load_ca_certificate(ca_crt)], crl_list)
     return crt, key, ca_crt
   # Latest error comes from validateCertAndKey
   raise # pylint: disable=misplaced-bare-raise
@@ -337,7 +371,7 @@ def validateCertAndKey(cert_pem, key_pem):
   ).public_key().public_numbers():
     raise ValueError('Mismatch between private key and certificate')
 
-def _verifyCertificateChain(cert, trusted_cert_list, crl):
+def _verifyCertificateChain(cert, trusted_cert_list, crl_list):
   """
   Verifies whether certificate has been signed by any of the trusted
   certificates, is not revoked and is whithin its validity period.
@@ -349,13 +383,18 @@ def _verifyCertificateChain(cert, trusted_cert_list, crl):
   # seems to be a rare implementation of it, so we keep using this module.
   # BUT it MUST NOT be used anywhere outside this function (hence the
   # bad-style local import). Use "cryptography".
+  # Also, older pylint (last version suppoting 2.7 ?) does not support
+  # import-outside-toplevel but does not detect anything wrong here.
+  # pylint: disable=import-outside-toplevel
   from OpenSSL import crypto
+  # pylint: enable=import-outside-toplevel
   store = crypto.X509Store()
   assert trusted_cert_list
   for trusted_cert in trusted_cert_list:
     store.add_cert(crypto.X509.from_cryptography(trusted_cert))
-  if crl is not None:
-    store.add_crl(crypto.CRL.from_cryptography(crl))
+  if crl_list:
+    for crl in crl_list:
+      store.add_crl(crypto.CRL.from_cryptography(crl))
     store.set_flags(crypto.X509StoreFlags.CRL_CHECK)
   try:
     crypto.X509StoreContext(
@@ -456,7 +495,7 @@ def load_ca_certificate(data):
   _verifyCertificateChain(crt, [crt], None)
   return crt
 
-def load_certificate(data, trusted_cert_list, crl):
+def load_certificate(data, trusted_cert_list, crl_list):
   """
   Load a certificate from PEM-encoded data.
 
@@ -464,7 +503,7 @@ def load_certificate(data, trusted_cert_list, crl):
   any of trusted certificates, is revoked or is otherwise invalid.
   """
   crt = x509.load_pem_x509_certificate(data, _cryptography_backend)
-  _verifyCertificateChain(crt, trusted_cert_list, crl)
+  _verifyCertificateChain(crt, trusted_cert_list, crl_list)
   return crt
 
 def dump_certificate(data):
@@ -534,6 +573,26 @@ def load_crl(data, trusted_cert_list):
       return crl
   raise cryptography.exceptions.InvalidSignature
 
+def _getAuthorityKeyIdentifier(cert):
+  return cert.extensions.get_extension_for_class(
+    x509.AuthorityKeyIdentifier,
+  ).value.key_identifier
+
+if sys.version_info < (3, ): # pragma: no cover
+  def getAuthorityKeyIdentifier(cert):
+    """
+    Returns the authority key identifier of given certificate.
+    """
+    return int(_getAuthorityKeyIdentifier(cert).encode('hex'), 16)
+else: # pragma: no cover
+  def getAuthorityKeyIdentifier(cert):
+    """
+    Returns the authority key identifier of given certificate.
+    """
+    # pylint: disable=no-member
+    return int.from_bytes(_getAuthorityKeyIdentifier(cert), 'big')
+    # pylint: enable=no-member
+
 EPOCH = datetime.datetime(1970, 1, 1)
 def datetime2timestamp(value):
   """
@@ -541,6 +600,12 @@ def datetime2timestamp(value):
   """
   return int((value - EPOCH).total_seconds())
 
+def timestamp2datetime(value):
+  """
+  Convert given unix timestamp into a datetime.
+  """
+  return EPOCH + datetime.timedelta(0, value)
+
 def timestamp2IMFfixdate(value):
   """
   Convert a timestamp into an IMF-fixdate string following RFC7231.
