@@ -36,12 +36,12 @@ import jwt
 from . import utils
 from . import exceptions
 
-# pylint: disable=import-error
+# pylint: disable=import-error,no-name-in-module
 if sys.version_info >= (3, ): # pragma: no cover
   from html import escape
 else: # pragma: no cover
   from cgi import escape
-# pylint: enable=import-error
+# pylint: enable=import-error,no-name-in-module
 
 __all__ = ('Application', 'CORSTokenManager')
 
@@ -172,9 +172,16 @@ class InsufficientStorage(ApplicationError):
   """
   No storage slot available (not necessarily out of disk space)
   """
-  # httplib lacks the textual description for 507, although it has the
-  # constant...
+  # python2.7's httplib lacks the textual description for 507, although it
+  # has the constant.
+  # And modern pylint on python3 complain that
+  # http.client.INSUFFICIENT_STORAGE, an enum item, is not suitable for %i
+  # (spoiler: it is suitable).
+  # Also, older pylint (last version suppoting 2.7 ?) does not support
+  # bad-string-format-type but does not detect anything wrong here.
+  # pylint: disable=bad-string-format-type
   status = '%i Insufficient Storage' % (httplib.INSUFFICIENT_STORAGE, )
+  # pylint: enable=bad-string-format-type
 
 STATUS_OK = _getStatus(httplib.OK)
 STATUS_CREATED = _getStatus(httplib.CREATED)
@@ -350,10 +357,24 @@ class Application(object):
         'method': {
           'GET': {
             'do': self.getCertificateRevocationList,
-            'descriptor': [{
-              'name': 'getCertificateRevocationList',
-              'title': 'Retrieve latest certificate revocation list.',
-            }],
+            'subpath': SUBPATH_OPTIONAL,
+            'descriptor': [
+              {
+                'name': 'getCertificateRevocationListList',
+                'title': (
+                  'Retrieve latest certificate revocation list for all valid '
+                  'authorities.'
+                ),
+              },
+              {
+                'name': 'getCertificateRevocationList',
+                'title': (
+                  'Retrieve latest certificate revocation list for given '
+                  'decimal representation of the authority identifier.'
+                ),
+                'subpath': '{+authority_key_id}',
+              },
+            ],
           },
         },
       },
@@ -529,7 +550,7 @@ class Application(object):
           except KeyError:
             break
           del path_item_list[0]
-        # If this raises, it means the outing dict is inconsistent.
+        # If this raises, it means the routing dict is inconsistent.
         method_dict = path_entry_dict['method']
         request_method = environ['REQUEST_METHOD']
         try:
@@ -631,8 +652,14 @@ class Application(object):
     Raises TooLarge if Content-Length if over MAX_BODY_LENGTH.
     If Content-Length is not set, reads at most MAX_BODY_LENGTH bytes.
     """
+    content_length = environ.get('CONTENT_LENGTH')
+    if not content_length:
+      result = environ['wsgi.input'].read(MAX_BODY_LENGTH)
+      if environ['wsgi.input'].read(1):
+        raise TooLarge(b'Content-Length limit exceeded')
+      return result
     try:
-      length = int(environ.get('CONTENT_LENGTH') or MAX_BODY_LENGTH)
+      length = int(content_length, 10)
     except ValueError:
       raise BadRequest(b'Invalid Content-Length')
     if length > MAX_BODY_LENGTH:
@@ -651,10 +678,10 @@ class Application(object):
       utils.load_certificate(
         environ.get('SSL_CLIENT_CERT', b''),
         trusted_cert_list=ca_list,
-        crl=utils.load_crl(
-          self._cau.getCertificateRevocationList(),
-          ca_list,
-        ),
+        crl_list=[
+          utils.load_crl(x, ca_list)
+          for x in self._cau.getCertificateRevocationListDict().itervalues()
+        ],
       )
     except (exceptions.CertificateVerificationError, ValueError):
       raise SSLUnauthorized
@@ -687,8 +714,8 @@ class Application(object):
     Returns a Morsel instance.
     """
     cookie = SimpleCookie({self._cors_cookie_id: value})[self._cors_cookie_id]
-    cookie['path'] = environ.get('SCRIPT_NAME') or '/',
-    cookie['expires'] = A_YEAR_IN_SECONDS
+    cookie['path'] = environ.get('SCRIPT_NAME') or '/'
+    cookie['max-age'] = A_YEAR_IN_SECONDS
     # No "secure" flag: cookie is not secret, and is protected against
     # tampering on client side.
     # No "httponly" flag: cookie is protected against tampering on client side,
@@ -876,10 +903,11 @@ class Application(object):
       'application/hal+json',
     )
 
-  def getCORSForm(self, context, environ): # pylint: disable=unused-argument
+  def getCORSForm(self, context, environ):
     """
     Handle GET /cors .
     """
+    _ = context # Silence pylint
     if environ['wsgi.url_scheme'] != 'https':
       return (
         STATUS_FOUND,
@@ -910,10 +938,11 @@ class Application(object):
       ],
     )
 
-  def postCORSForm(self, context, environ): # pylint: disable=unused-argument
+  def postCORSForm(self, context, environ):
     """
     Handle POST /cors .
     """
+    _ = context # Silence pylint
     if environ['wsgi.url_scheme'] != 'https':
       raise NotFound
     if environ.get('CONTENT_TYPE') != 'application/x-www-form-urlencoded':
@@ -954,18 +983,25 @@ class Application(object):
       [],
     )
 
-  def getCertificateRevocationList(
-    self,
-    context,
-    environ,
-  ): # pylint: disable=unused-argument
+  def getCertificateRevocationList(self, context, environ, subpath):
     """
-    Handle GET /{context}/crl .
+    Handle GET /{context}/crl and GET /{context}/crl/{authority_key_id} .
     """
-    return self._returnFile(
-      context.getCertificateRevocationList(),
-      'application/pkix-crl',
-    )
+    _ = environ # Silence pylint
+    crl_dict = context.getCertificateRevocationListDict()
+    if subpath:
+      try:
+        authority_key_id, = subpath
+        authority_key_id = int(authority_key_id, 10)
+      except ValueError:
+        raise NotFound
+      try:
+        crl = crl_dict[authority_key_id]
+      except KeyError:
+        raise NotFound
+    else:
+      crl = b'\n'.join(crl_dict.itervalues())
+    return self._returnFile(crl, 'application/pkix-crl')
 
   def getCSR(self, context, environ, subpath):
     """
@@ -1009,41 +1045,31 @@ class Application(object):
       raise NotFound
     return (STATUS_NO_CONTENT, header_list, [])
 
-  def getCACertificate(
-    self,
-    context,
-    environ,
-  ): # pylint: disable=unused-argument
+  def getCACertificate(self, context, environ):
     """
     Handle GET /{context}/crt/ca.crt.pem urls.
     """
+    _ = environ # Silence pylint
     return self._returnFile(
       context.getCACertificate(),
       'application/x-x509-ca-cert',
     )
 
-  def getCACertificateChain(
-    self,
-    context,
-    environ,
-  ): # pylint: disable=unused-argument
+  def getCACertificateChain(self, context, environ):
     """
     Handle GET /{context}/crt/ca.crt.json urls.
     """
+    _ = environ # Silence pylint
     return self._returnFile(
       json.dumps(context.getValidCACertificateChain()).encode('utf-8'),
       'application/json',
     )
 
-  def getCertificate(
-    self,
-    context,
-    environ,
-    subpath,
-  ): # pylint: disable=unused-argument
+  def getCertificate(self, context, environ, subpath):
     """
     Handle GET /{context}/crt/{crt_id} urls.
     """
+    _ = environ # Silence pylint
     return self._returnFile(
       context.getCertificate(self._getCSRID(subpath)),
       'application/pkix-cert',
